/*
 * ITELearner.cpp
 *
 *  Created on: Jan 14, 2014
 *      Author: nghiaht
 */

#include "ITELearner.h"

ITELearner::ITELearner(Mat &Field, int N, vector <IsoGaussian*> &covfunc, int nL, Vec &b)
{
	this->Field = Field; this->N = N; this->nL = nL;
	this->covfunc = covfunc; this->b = b; this->hmap =  Vec(N, -1);

	marker = new bool[N];
	for (int i = 0; i < N; i++) marker[i] = false;

	this->m = 0; this->X.clear(); this->Y.clear();
	this->K = vector <Mat> (this->nL); this->L = vector <Mat>(this->nL);

	for (int t = 0; t < this->nL; t++)
	{
		this->K[t].clear();
		this->L[t].clear();
	}

	this->left = 0; this->right = nL - 1; this->htest = Vec(0);
}

Vec ITELearner::Move(int n)
{
	n = min(n, this->N - this->m);

	Vec list = Vec(0); this->htest.clear();

	if ((this->left < this->right) && (n > 1))
	{
		int nSample = min(N_SAMPLE, n), mid = (this->left + this->right) / 2, pos;
		double l1 = this->covfunc[mid]->Bandwidth(), l2 = this->covfunc[mid + 1]->Bandwidth();
		double delta = ceil(2 * l1 * l2  * sqrt(log(l2 / l1) / (l2 * l2 - l1 * l1)));

		for (int i = 0; i < nSample; i++)
		{
			pos = (int) (i * delta);
			if (pos > this->N - 1) break;

			this->htest.push_back(pos);
			if (!this->marker[pos]) list.push_back(pos);
		}
	}
	else list = AprioriDesign(n, this->left);

	return list;
}

void ITELearner::Update(Vec selected, Vec obs)
{
	// marking new observations
	for (int i = 0; i < (int) selected.size(); i++)
	{
		this->marker[(int) selected[i]] = true;
		this->hmap[(int) selected[i]] = obs[i];
		this->X.push_back(selected[i]); this->Y.push_back(obs[i]);

		for (int t = 0; t < this->nL; t++)
		{
			for (int j = 0; j < this->m; j++) this->K[t][j].push_back(this->covfunc[t]->Eval(this->Field[(int) selected[i]], this->Field[this->X[j]]));
			this->K[t].push_back(Vec(this->m + 1, 0.0));

			for (int j = 0; j < this->m; j++) this->K[t][this->m][j] = this->K[t][j][this->m];
			this->K[t][this->m][this->m] = this->covfunc[t]->Var(this->Field[(int) selected[i]]);

			this->L[t] = Cholesky(this->K[t]);
		}

		this->m = this->m + 1;
	}

	if ((this->left < this->right) && ((int) this->htest.size() > 1)) // run the latest hypothesis test: computing the Pearson correlation coefficient
	{
		int nSample = (int) this->htest.size(), mid = (this->left + this->right) / 2;
		double pearson = 0.0, x_mean = 0.0, y_mean = 0.0;

		for (int i = 0; i < nSample; i++)
		{
			x_mean += this->htest[i];
			y_mean += this->hmap[this->htest[i]];
		}

		x_mean = (x_mean * 1.0) / nSample; y_mean = (y_mean * 1.0) / nSample;

		double numerator = 0.0, dX = 0.0, dY = 0.0;

		for (int i = 0; i < nSample; i++)
		{
			numerator += (this->htest[i] - x_mean) * (this->hmap[this->htest[i]] - y_mean);
			dX += (this->htest[i] - x_mean) * (this->htest[i] - x_mean);
			dY += (this->hmap[this->htest[i]] - y_mean) * (this->hmap[this->htest[i]] - y_mean);
		}

		pearson = numerator / (dX * dY);
		if (fabs(pearson) > THRESHOLD) this->left = mid + 1; else this->right = mid;
	}
}

double ITELearner::Entropy(int s, int kernel)
{
	double var = this->covfunc[kernel]->Var(this->Field[s]);

	if (this->m > 0)
	{
		Vec k(this->m, 0.0);
		for (int j = 0; j < this->m; j++) k[j] = this->covfunc[kernel]->Eval(this->Field[s], this->Field[X[j]]);
		Vec v = ForwardSub(this->L[kernel], k);
		var = var - Dot(v, v);
	}

	return ENT(var);
}

int ITELearner::GreedyChoice(int kernel)
{
	int res = -1; double best_score = -INFTY;

	for (int s = 0; s < this->N; s++)
		if (!this->marker[s])
		{
			double entropy = Entropy(s, kernel);
			if (entropy > best_score)
			{
				best_score = entropy;
				res = s;
			}
		}

	return res;
}

Vec ITELearner::AprioriDesign(int n, int kernel)
{
	n = min(n, this->N - this->m);
	Vec res(n, -1);

	for (int i = 0; i < n; i++)
	{
		int next = GreedyChoice(kernel);
		res[i] = next;

		this->marker[next] = true; this->X.push_back(next);

		for (int j = 0; j < this->m; j++) this->K[kernel][j].push_back(this->covfunc[kernel]->Eval(this->Field[next], this->Field[X[j]]));
		this->K[kernel].push_back(Vec(this->m + 1, 0.0));
		for (int j = 0; j < this->m; j++) this->K[kernel][this->m][j] = this->K[kernel][j][this->m];
		this->K[kernel][this->m][this->m] = this->covfunc[kernel]->Var(this->Field[next]); this->m = this->m + 1;

		this->L[kernel] = Cholesky(this->K[kernel]);
	}

	return res;
}




