/*
 * Node.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "Node.h"

Node::Node(int n, int dx, double dy, Node* parent, Factory* company, Vec &b, double LB, double UB)
{
	this->n = n; this->dx = dx; this->dy = dy;
	this->boss = parent; this->company = company; this->b = b;
	this->selected = -1; this->LB = LB; this->UB = UB;
	this->subordinates.clear(); this->dCost.clear();

	if (this->n == 0)
	{
		this->LB = 0.0;
		this->UB = 0.0;
	}

	if (this->n == 1) BaseCase();
}

void Node::BaseCase()
{
	Vec X(0); Vec Y(0);

	this->company->GetPriorData(X, Y);

	Trace(this, X, Y);

	vector < vector < pair <double, double> > > pGP = this->company->PosteriorGP(X, Y);
	vector < pair <int, Vec> > samples = this->company->Sample(X, pGP, this->b, NSAMPLE);

	double cost = 0.0, costMax = -INFTY; this->selected = samples[0].first;

	for (int i = 0; i < (int) samples.size(); i++)
	{
		cost = 0.0;

		for (int s = 0; s < NSAMPLE; s++)
		{
			double z = samples[i].second[s], sum = 0.0;
			for (int t = 0; t < (int) this->b.size(); t++) sum += this->b[t] * N(z, pGP[i][t].first, pGP[i][t].second);
			cost -= log(sum);
		}

		cost /= NSAMPLE;

		if (cost > costMax)
		{
			this->selected = samples[i].first;
			costMax = cost;
		}
	}

	this->LB = costMax; this->UB = costMax;
}

Node::Node(int n, int dx, double dy, Node* parent, Factory* company, Vec &b)
{
	this->n = n; this->dx = dx; this->dy = dy;
	this->boss = parent; this->company = company; this->b = b;
	this->selected = -1; this->LB = -INFTY; this->UB = INFTY;
	this->subordinates.clear(); this->dCost.clear();
	this->InitializeBounds();
}

Node::~Node()
{
	for (int i = 0; i < (int) this->subordinates.size(); i++)
		delete this->subordinates[i];

	this->subordinates.clear();
}

void Node::InitializeBounds()
{
	Vec X(0); Vec Y(0);

	this->company->GetPriorData(X, Y);

	Trace(this, X, Y);

	pair <double, double> bounds = this->company->InitialBounds(X, this->b, this->n);

	this->LB = bounds.first - BIAS; this->UB = bounds.second + BIAS;
}

void Node::Trace(Node *current, Vec &X, Vec &Y)
{
	if (current->boss == NULL) return;

	Trace(current->boss, X, Y);

	pair <int, double> last = current->LastObservation();
	X.push_back(last.first); Y.push_back(last.second);
}

int Node::Budget()
{
	return this->n;
}

vector <Node*> Node::Report()
{
	return this->subordinates;
}

int Node::NextLocation()
{
	return this->selected;
}

pair <int, double> Node::LastObservation()
{
	return make_pair(this->dx, this->dy);
}

pair <double, double> Node::GetBounds()
{
	return make_pair(this->LB, this->UB);
}

Node* Node::Boss()
{
	return this->boss;
}

double Node::BoundGap()
{
	return fabs(this->UB - this->LB);
}

void Node::Explore()
{
	Vec X(0); Vec Y(0);

	this->company->GetPriorData(X, Y);

	Trace(this, X, Y);

	vector < vector < pair <double, double> > > pGP = this->company->PosteriorGP(X, Y);
	vector < pair <int, Vec> > samples = this->company->Sample(X, pGP, this->b, NSAMPLE);

	X.push_back(-1); // reserve a new place for the yet-to-be selected sampling site
	double offset = this->company->SubmodularConstant();
	int nL = (int) this->b.size();

	for (int i = 0; i < (int) samples.size(); i++)
	{
		int site = samples[i].first;
		X[(int) X.size() - 1] = site;

		double newLB = 0.0, newUB = 0.0, cost = 0.0;

		if (this->n > 2) newLB = this->company->GreedyDesign(X, this->n - 1, this->b).first;

		Vec nUB(nL, 0.0);
		if (this->n > 2)
			for (int t = 0; t < nL; t++) nUB[t] = this->company->GreedyDesign(X, this->n - 1, t).first;

		for (int s = 0; s < NSAMPLE; s++)
		{
			double z = samples[i].second[s], sum = 0.0;
			Vec pb(nL, 0.0); newUB = 0.0;

			for (int t = 0; t < nL; t++)
			{
				pb[t] = b[t] * N(z, pGP[i][t].first, pGP[i][t].second);
				sum += pb[t];
			}

			cost -= log(sum);

			for (int t = 0; t < nL; t++)
			{
				pb[t] /= sum;
				if (this->n > 1 && pb[t] > EPS) newUB -= log(pb[t]) * pb[t];
				newUB += pb[t] * ((E * nUB[t] + (this->n - 1) * offset) / (E - 1));
			}

			Node* child = new Node(this->n - 1, site, z, this, this->company, pb, newLB - BIAS, newUB + BIAS);
			this->subordinates.push_back(child);
		}

		this->dCost.push_back(cost / NSAMPLE);
	}
}

void Node::UpdateBounds()
{
	int nSite = (int) this->dCost.size(), index = 0;
	double nLB = -INFTY, nUB = -INFTY;

	if (nSite == 0)
	{
		nLB = 0;
		nUB = 0;
	}

	for (int s = 0; s < nSite; s++)
	{
		double cost_lb = this->dCost[s], cost_ub = this->dCost[s];
		for (int i = 0; i < NSAMPLE; i++)
		{
			pair <double, double> bounds = this->subordinates[index]->GetBounds();
			cost_lb += (bounds.first / NSAMPLE);
			cost_ub += (bounds.second / NSAMPLE);
			index++;
		}

		if (cost_lb > nLB) this->selected = this->subordinates[index - 1]->LastObservation().first;
		nLB = max(nLB, cost_lb); nUB = max(nUB, cost_ub);
	}

	this->UB = min(this->UB, nUB); this->LB = max(this->LB, nLB);

	if (nSite > 0 && this->selected < 0) cout << "WHY" << endl;

	if (this->boss != NULL) this->boss->UpdateBounds();
}



