/*
 * utility.cpp
 *
 *  Created on: Oct 26, 2013
 *      Author: nghiaht
 */

#include "utility.h"

Mat Cholesky(Mat &A) // Cholesky Decomposition -- Only applicable to positive definite matrices
{
    int N = (int) A.size();
    Mat L(N, Vec(N, 0.0));

    double sum = 0.0;

    for (int j = 0; j < N; j++)
    {
        sum = 0.0;
        for (int k = 0; k < j; k++)
            sum += L[j][k] * L[j][k];
        L[j][j] = sqrt(A[j][j] - sum);

        for (int i = j + 1; i < N; i++)
        {
            sum = 0.0;
            for (int k = 0; k < j; k++)
                sum += L[i][k] * L[j][k];

            L[i][j] = (A[i][j] - sum) / L[j][j];
        }
    }

    return L;
}

Vec ForwardSub(Mat &A, Vec &b) // solving Ax = b -- For lower triangular matrix
{
    int N = (int) A.size();

    Vec X(N, 0.0);

    X[0] = b[0] / A[0][0];
    double sum = 0.0;


    for (int i = 1; i < N; i++)
    {
        sum = 0.0;
        for (int k = 0; k < i; k++)
            sum += X[k] * A[i][k];

        X[i] = (b[i] - sum) / A[i][i];
    }

    return X;
}

Vec BackwardSub(Mat &A, Vec &b) // solving Ax = b -- For upper triangular matrix
{
    int N = (int) A.size();
    Vec X(N, 0.0); X[N - 1] = b[N - 1] / A[N - 1][N - 1];
    double sum = 0.0;

    for (int i = N - 2; i >= 0; i--)
    {
        sum = 0.0;
        for (int k = i + 1; k < N; k++)
            sum += X[k] * A[i][k];
        X[i] = (b[i] - sum) / A[i][i];
    }

    return X;
}

double Dot(Vec &a, Vec &b) // dot product of two vectors
{
    int N = (int) a.size();
    double res = 0.0;
    for (int i = 0; i < N; i++)
        res += a[i] * b[i];
    return res;
}

Vec RanVec(int N, double a, double b, bool integral) // generate a random vector whose components are independently drawn from a uniform distribution -- the integral flag indicates whether we want integer or real components
{
    Vec X(N, 0);

    for (int i = 0; i < N; i++)
    {
        if (integral) X[i] = (int) IRAND((int) a, (int) b);
        else X[i] = DRAND(a, b);
    }

    return X;
}

Mat RanMat(int M, int N, double a, double b, bool integral) // generate a random matrix whose components are independently drawn from a uniform distribution
{
    Mat X;

    for (int i = 0; i < M; i++) X.push_back(RanVec(N, a, b, integral));

    return X;
}

Mat Mult(Mat &A, Mat &B) // multiply two matrices who have matching size -- for example, A(m,n) 8 B(n, k)
{
    int M = (int) A.size(), K = (int) A[0].size(), N = (int) B[0].size();

    Mat X(M, Vec(N, 0.0));

    for (int i = 0; i < M; i++)
        for (int j = 0; j < N; j++)
            for (int k = 0; k < K; k++) X[i][j] += A[i][k] * B[k][j];

    return X;
}

Mat Trans(Mat &A) // transpose a matrix
{
    int M = (int) A.size(), N = (int) A[0].size();

    Mat B(N, Vec(N, 0));

    for (int i = 0; i < N; i++)
        for (int j = 0; j < M; j++) B[i][j] = A[j][i];

    return B;
}

double grand() // Marsaglia polar method -- generate a sample from N(0, 1)
{
    double u, v, s = 2.0;

    while ((s >= 1.0) || (s == 0.0))
    {
        u = DRAND(-1, 1);
        v = DRAND(-1, 1);
        s = u * u + v * v;
    }

    return u * sqrt(-2 * log(s) / s);
}

Vec MultiGaussRan(Vec &mean, Mat &cov) // generate a sample from N(mean, cov)
{
    Mat L = Cholesky(cov);
    int N = (int) mean.size();

    Vec Z(N, 0);
    for (int i = 0; i < N; i++) Z[i] = GRAND(0, 1);

    Vec X(N, 0);
    for (int i = 0; i < N; i++)
    {
        X[i] = mean[i];
        for (int j = 0; j < N; j++) X[i] += (L[i][j] * Z[j]);
    }

    return X;
}

Vec grand(double mean, double sigma, int N) // generate N samples independently from N(mean, sigma^2)
{
    Vec X(N, 0);
    for (int i = 0; i < N; i++)
        X[i] = GRAND(mean, sigma);
    return X;
}

Mat MultiGaussRan(Vec &mean, Mat &cov, int N) // generate N samples independently from N(mean, cov)
{
    Mat X;
    for (int i = 0; i < N; i++)
        X.push_back(MultiGaussRan(mean, cov));
    return X;
}

int DiscreteSample(Vec &b) // generate a sample from a discrete distribution b
{
	int selected = (int) b.size() - 1;
	double chance = DRAND(0, 1), sum = 0.0;

	for (int t = 0; t < (int) b.size(); t++)
	{
		sum += b[t];
		if (sum >= chance) return t;
	}

	return selected;
}


